@@ -36,6 +36,8 @@ import json
 from urllib.parse import urlparse
 import string
 import random
+import traceback
+import hashlib
 
 # import astpretty
 
@@ -46,8 +48,12 @@ TRANSFORM_CSV_DS = "infin_transform_csv_to_ds"
 TRANSFORM_ONE_OBJ = "infin_transform_one_object"
 TRANSFORM_DIR_BY_DIR = "infin_transform_dir_by_dir"
 TRANSFORM_ALL_OBJECTS = "infin_transform_all_objects"
+TRANSFORM_MULTI_INPUTS = "infin_transform_all_objects_multi_inputs"
 
-verbose = False
+# conda env config vars set INFINSTOR_VERBOSE=True : to set the environment variable
+# conda env config vars unset  : to unset the environment variable
+# conda env config vars list  : to list the environment variables
+verbose = bool(os.getenv('INFINSTOR_VERBOSE'))
 enable_infin_ast = False # right now infin_ast does not pick up classes defined in the cell
 
 def num_threads():
@@ -322,7 +328,7 @@ def s3downloader(**kwargs):
                 except Exception as e:
                     info('s3downloader: Caught ' + str(e) + ' downloading s3://'
                             + str(bucketname) + '/' + str(full_object_key)
-                            + ' to ' + str(tmpf_name))
+                            + ' to ' + str(tmpf_name) + ", traceback = \n" + traceback.format_exc())
 
                 content_length = os.stat(tmpf_name).st_size
                 key = bucketname + '/' + full_object_key
@@ -340,7 +346,7 @@ def s3downloader(**kwargs):
                         write_pipe.send('Success')
                 except Exception as e:
                     status_str = str(e)
-                    info("Error sending bytes back: " + status_str)
+                    info("Error sending bytes back: " + status_str + ", traceback = \n" + traceback.format_exc())
             else:
                 try:
                     temp_output_dir = tempfile.mkdtemp()
@@ -350,7 +356,7 @@ def s3downloader(**kwargs):
                             parentdir[len(prefix_trunc):], **kwargs)
                 except Exception as e1:
                     status_str = str(e1)
-                    info("Error executing infin_transform_one_object_tmpfile: " + status_str)
+                    info("Error executing infin_transform_one_object_tmpfile: " + status_str + ", traceback = \n" + traceback.format_exc())
                 else:
                     log_artifacts_recursively(temp_output_dir)
                     status_str = 'Success'
@@ -505,7 +511,7 @@ def actually_run_transformation(client, is_pandas_df, bucketname,\
         exec(compiledcode2, glb)
     except Exception as e:
         status_str = str(e)
-        info("Execution of global statics failed. status_str=" + status_str)
+        info("Execution of global statics failed. status_str=" + status_str + ", traceback = \n" + traceback.format_exc())
         raise
     else:
         status_str = 'Success'
@@ -518,7 +524,7 @@ def actually_run_transformation(client, is_pandas_df, bucketname,\
         namespaced_infin_transform_fnx(objects, **kwargs)
     except Exception as e:
         status_str = str(e)
-        info("Error executing " + xn + ", status=" + status_str)
+        info("Error executing " + xn + ", status=" + status_str + ", traceback = \n" + traceback.format_exc())
     else:
         status_str = 'Success'
 
@@ -549,7 +555,7 @@ def run_script(transformation_string, **kwargs):
         exec(compiledcode2, glb)
     except Exception as e:
         status_str = str(e)
-        info("Execution of global statics failed. status_str=" + status_str)
+        info("Execution of global statics failed. status_str=" + status_str + ", traceback = \n" + traceback.format_exc())
         raise
     else:
         status_str = 'Success'
@@ -605,6 +611,7 @@ def read_and_xform_one_object(client, bucketname,\
     if (verbose == True):
         print('read_and_xform_one_object: num entries in dict_of_arrays_of_files='\
             + str(len(dict_of_arrays_of_files)))
+    objects = None
     for parentdir in dict_of_arrays_of_files:
         array_of_files = dict_of_arrays_of_files[parentdir]
         info('Number Of Objects in parentdir ' + parentdir + ' = ' + str(len(array_of_files)))
@@ -617,7 +624,7 @@ def read_and_xform_one_object(client, bucketname,\
         except Exception as e:
             status_str = str(e)
             info("Execution of global statics failed for " + parentdir\
-                    + ". status_str=" + status_str)
+                    + ". status_str=" + status_str + ", traceback = \n" + traceback.format_exc())
             raise
         else:
             status_str = 'Success'
@@ -676,7 +683,7 @@ def read_and_xform_dir_by_dir(client, bucketname, prefix_in, prefix_trunc, trans
             status_str = str(e)
             shutil.rmtree(temp_input_dir)
             info("Execution of global statics failed for " + parentdir\
-                    + ". status_str=" + status_str)
+                    + ". status_str=" + status_str + ", traceback = \n" + traceback.format_exc())
             continue # process next directory
         else:
             status_str = 'Success'
@@ -689,14 +696,15 @@ def read_and_xform_dir_by_dir(client, bucketname, prefix_in, prefix_trunc, trans
             namespaced_infin_transform_fnx(temp_input_dir, temp_output_dir, parentdir[len(prefix_trunc):], **kwargs)
         except Exception as e:
             status_str = str(e)
-            info("Error executing " + xn + ", status=" + status_str)
+            info("Error executing " + xn + ", status=" + status_str + ", traceback = \n" + traceback.format_exc())
         else:
             info("Successfully executed transform " + xn + " for parentdir " + parentdir[len(prefix_trunc):])
         finally:
             shutil.rmtree(temp_input_dir)
 
         try:
-            dest_path = 'infinstor/' + parentdir[len(prefix_trunc):]
+            # lstrip() ensures that parentdir[....] expression does not start with a '/'.  this is needed to ensure that we don't have a '//' like 'infinstor//...'
+            dest_path = 'infinstor/' + parentdir[len(prefix_trunc):].lstrip('/')  
             for one_output_file in os.listdir(temp_output_dir):
                 fq_local = os.path.join(temp_output_dir, one_output_file)
                 info("Logging artifact " + str(fq_local) + " to " + dest_path)
@@ -704,7 +712,7 @@ def read_and_xform_dir_by_dir(client, bucketname, prefix_in, prefix_trunc, trans
         except Exception as e:
             status_str = str(e)
             info("Error logging artifacts for parentdir " + str(parentdir)\
-                    + ": " + status_str)
+                    + ": " + status_str + ", traceback = \n" + traceback.format_exc())
         else:
             info("Successfully logged artifacts for parentdir " + str(parentdir))
         finally:
@@ -758,7 +766,7 @@ def read_and_xform_all_objects(client, bucketname, prefix_in, prefix_trunc, tran
     except Exception as e:
         status_str = str(e)
         shutil.rmtree(temp_input_dir_root)
-        info("Execution of global statics failed. status_str=" + status_str)
+        info("Execution of global statics failed. status_str=" + status_str + ", traceback = \n" + traceback.format_exc())
         raise
     else:
         status_str = 'Success'
@@ -771,21 +779,22 @@ def read_and_xform_all_objects(client, bucketname, prefix_in, prefix_trunc, tran
         namespaced_infin_transform_fnx(temp_input_dir_root, temp_output_dir, **kwargs)
     except Exception as e:
         status_str = str(e)
-        info("Error executing " + xn + ", status=" + status_str)
+        info("Error executing " + xn + ", status=" + status_str + ", traceback = \n" + traceback.format_exc())
     else:
         info("Successfully executed transform " + xn)
     finally:
         shutil.rmtree(temp_input_dir)
 
     try:
-        dest_path = 'infinstor/' + prefix_in[len(prefix_trunc):]
+        # lstrip() ensures that prefix_in[...] expression does not start with a '/'.  this is needed to ensure that we don't have a '//' like 'infinstor//...'
+        dest_path = 'infinstor/' + prefix_in[len(prefix_trunc):].lstrip('/') 
         for one_output_file in os.listdir(temp_output_dir):
             fq_local = os.path.join(temp_output_dir, one_output_file)
             info("Logging artifact " + str(fq_local) + " to " + dest_path)
             log_artifact(fq_local, dest_path)
     except Exception as e:
         status_str = str(e)
-        info("Error logging artifacts: " + status_str)
+        info("Error logging artifacts: " + status_str + ", traceback = \n" + traceback.format_exc())
     else:
         info("Successfully logged artifacts")
     finally:
@@ -811,34 +820,38 @@ def look_for_transform(transform_string, transform_symbol):
 def get_mlflow_run_artifacts_info(run_id):
     client = mlflow.tracking.MlflowClient()
     run = client.get_run(run_id)
-    artifact_uri = run.info.artifact_uri
+    # run.info = <RunInfo: artifact_uri='s3://infinstor-mlflow-artifacts-rrajendran-isstage2/mlflow-artifacts/rrajendran/0/0-a3eae5ebd76b4f35bb0cf1ef0162564f', end_time=1613664747468, experiment_id='0', lifecycle_stage='active', run_id='0-a3eae5ebd76b4f35bb0cf1ef0162564f', run_uuid='0-a3eae5ebd76b4f35bb0cf1ef0162564f', start_time=1613664685420, status='FINISHED', user_id='rrajendran'>
+    artifact_uri = run.info.artifact_uri    
     parse_result = urlparse(artifact_uri)
     if (parse_result.scheme != 's3'):
         raise ValueError('Error. Do not know how to deal with artifacts in scheme '\
                 + parse_result.scheme)
     bucketname = parse_result.netloc
-    prefix = parse_result.path.lstrip('/') + '/'
+    prefix = parse_result.path.lstrip('/') + '/'     # parse_result.path= '/mlflow-artifacts/rrajendran/0/0-a3eae5ebd76b4f35bb0cf1ef0162564f' ; prefix='mlflow-artifacts/rrajendran/0/0-a3eae5ebd76b4f35bb0cf1ef0162564f/'
     return bucketname, prefix
 
 def generate_cache_key(input_data_spec, xformname):
-    if (input_data_spec['type'] == 'infinsnap'):
-        cache_key = 'xform/' + xformname + '/infinsnap/'\
-                + input_data_spec['time_spec'] + '/' + input_data_spec['bucketname']\
-                + '/' + input_data_spec['prefix']
-    elif (input_data_spec['type'] == 'infinslice'):
-        cache_key = 'xform/' + xformname + '/infinslice/'\
-                + input_data_spec['time_spec'] + '/' + input_data_spec['bucketname']\
-                + '/' + input_data_spec['prefix']
-    elif (input_data_spec['type'] == 'label'):
-        cache_key = 'xform/' + xformname + '/label/' + input_data_spec['label']
-    elif (input_data_spec['type'] == 'mlflow-run-artifacts'):
-        cache_key = 'xform/' + xformname + '/run_id/' + input_data_spec['run_id']
-    elif (input_data_spec['type'] == 'no-input-data'):
-        cache_key = 'xform/' + xformname + '/no-input-data/' \
-            +''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(8))
-    else:
-        raise ValueError('Error. Unknown input_data_spec type ' + input_data_spec['type'])
-    return cache_key
+    input_json_str = json.dumps(input_data_spec)
+    input_json_bytes = input_json_str.encode('utf-8')
+    return hashlib.md5(input_json_bytes).hexdigest()
+    # if (input_data_spec['type'] == 'infinsnap'):
+    #     cache_key = 'xform/' + xformname + '/infinsnap/'\
+    #             + input_data_spec['time_spec'] + '/' + input_data_spec['bucketname']\
+    #             + '/' + input_data_spec['prefix']
+    # elif (input_data_spec['type'] == 'infinslice'):
+    #     cache_key = 'xform/' + xformname + '/infinslice/'\
+    #             + input_data_spec['time_spec'] + '/' + input_data_spec['bucketname']\
+    #             + '/' + input_data_spec['prefix']
+    # elif (input_data_spec['type'] == 'label'):
+    #     cache_key = 'xform/' + xformname + '/label/' + input_data_spec['label']
+    # elif (input_data_spec['type'] == 'mlflow-run-artifacts'):
+    #     cache_key = 'xform/' + xformname + '/run_id/' + input_data_spec['run_id']
+    # elif (input_data_spec['type'] == 'no-input-data'):
+    #     cache_key = 'xform/' + xformname + '/no-input-data/' \
+    #         +''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(8))
+    # else:
+    #     raise ValueError('Error. Unknown input_data_spec type ' + input_data_spec['type'])
+    # return cache_key
 
 def generate_input_data_spec_string(input_data_spec):
     if (input_data_spec['type'] == 'infinsnap'):
@@ -871,24 +884,29 @@ def check_in_cache(input_data_spec, xformname):
             if (run.info.status == 'FINISHED'):
                 return True, run.info.run_id
         return False, None
-    
 
-def run_transform_inline(service_name, run_id, input_data_spec, xformname, **kwargs):
+def get_data_connection_details(service_name, input_data_spec, xformname):
     cache_key = generate_cache_key(input_data_spec, xformname)
+    endpoint = None
+    prefix_in = None
+    prefix_trunc = ''
+    bucketname = None
+    session = None
+    client = None
     if (input_data_spec['type'] == 'infinsnap' or input_data_spec['type'] == 'infinslice'):
-        endpoint = "https://" + input_data_spec['time_spec'] + ".s3proxy."\
-                + service_name + ":443/";
+        endpoint = "https://" + input_data_spec['time_spec'] + ".s3proxy." \
+                   + service_name + ":443/";
         prefix_in = input_data_spec['prefix']
         prefix_trunc = ''
         bucketname = input_data_spec['bucketname']
         session = boto3.session.Session(profile_name='infinstor')
         client = session.client('s3', endpoint_url=endpoint)
     elif (input_data_spec['type'] == 'label'):
-        endpoint = "https://lbl" + input_data_spec['label'] + ".s3proxy."\
-                + service_name + ":443/";
-        prefix_in = '' # ignored for labels
-        prefix_trunc = '' # ignored for labels
-        bucketname = '' # ignored for labels
+        endpoint = "https://lbl" + input_data_spec['label'] + ".s3proxy." \
+                   + service_name + ":443/";
+        prefix_in = ''  # ignored for labels
+        prefix_trunc = ''  # ignored for labels
+        bucketname = ''  # ignored for labels
         session = boto3.session.Session(profile_name='infinstor')
         client = session.client('s3', endpoint_url=endpoint)
     elif (input_data_spec['type'] == 'mlflow-run-artifacts'):
@@ -906,17 +924,148 @@ def run_transform_inline(service_name, run_id, input_data_spec, xformname, **kwa
     else:
         raise ValueError('Error. Unknown input_data_spec type ' + input_data_spec['type'])
 
+    input_spec_string = generate_input_data_spec_string(input_data_spec)
+
+    return cache_key, endpoint, prefix_in, prefix_trunc, \
+           bucketname, session, client, input_spec_string
+
+
+def run_transform_inline(service_name, run_id, input_data_spec, xformname, **kwargs):
     transform_string, dockerfile = get_xform_info(xformname)
     if (transform_string == None):
         raise ValueError('Error. Cannot find xform ' + xformname)
-
     log_param('run_id', run_id)
-    log_param('cache_key', cache_key)
-    log_param('input_data_spec', generate_input_data_spec_string(input_data_spec))
-    xobjects = run_transform_string_inline(client, bucketname, prefix_in, prefix_trunc,
+
+    if type(input_data_spec) == list:
+        input_spec_string_list = []
+        clients = []
+        bucketnames = []
+        prefix_ins = []
+        prefix_truncs = []
+        for spec in input_data_spec:
+            cache_key, endpoint, prefix_in, prefix_trunc, bucketname, session, client, input_spec_string\
+                = get_data_connection_details(service_name, spec, xformname)
+            log_param('cache_key', cache_key)
+            input_spec_string_list.append(input_spec_string)
+            clients.append(client)
+            bucketnames.append(bucketname)
+            prefix_ins.append(prefix_in)
+            prefix_truncs.append(prefix_trunc)
+        log_param('input_data_spec', json.dumps(input_spec_string_list))
+        xobjects = run_transform_string_inline_multi_input(clients, bucketnames, prefix_ins, prefix_truncs,
+                    transform_string, **kwargs)
+    else:
+        cache_key, endpoint, prefix_in, prefix_trunc, bucketname, session, client, input_spec_string\
+            = get_data_connection_details(service_name, input_data_spec, xformname)
+        log_param('cache_key', cache_key)
+        log_param('input_data_spec', input_spec_string)
+        xobjects = run_transform_string_inline(client, bucketname, prefix_in, prefix_trunc,
                 transform_string, **kwargs)
+
     return xobjects
 
+def run_transform_string_inline_multi_input(clients, bucketnames, prefix_ins, prefix_truncs,
+                    transform_string, **kwargs):
+    if (enable_infin_ast == True):
+        transformAst = infin_ast.extract_transform(TRANSFORM_MULTI_INPUTS, src_str=transform_string)
+        transformSrc = infin_ast.get_source(transformAst)
+    else:
+        transformSrc = transform_string
+    tree = ast.parse(transformSrc)
+    # astpretty.pprint(tree)
+    compiledcode1 = compile(tree, "<string>", "exec")
+
+    temp_output_dir = tempfile.mkdtemp()
+    temp_input_roots = list()
+    total_dirs_to_process = 0
+    for i in range(len(clients)):
+        client = clients[i]
+        bucketname = bucketnames[i]
+        prefix_in = prefix_ins[i]
+        prefix_trunc = prefix_truncs[i]
+
+        dict_of_arrays_of_files = dict()
+        list_dir_by_dir(client, bucketname, prefix_in, True, dict_of_arrays_of_files)
+        if (len(dict_of_arrays_of_files) == 0):
+            info('No directories to process')
+            continue
+        total_dirs_to_process = total_dirs_to_process + len(dict_of_arrays_of_files)
+
+        temp_input_dir_root = tempfile.mkdtemp()
+        temp_input_roots.append(temp_input_dir_root)
+
+        for parentdir in dict_of_arrays_of_files:
+            temp_input_dir = os.path.join(temp_input_dir_root, parentdir[len(prefix_trunc):])
+            os.makedirs(temp_input_dir, mode=0o755, exist_ok=True)
+            print("DEBUG, created dir "+temp_input_dir)
+            array_of_files = dict_of_arrays_of_files[parentdir]
+            info('Processing parentdir ' + parentdir + ' with ' + str(len(array_of_files))\
+                    + ' objects, temp_input_dir=' + str(temp_input_dir)\
+                    + ', temp_output_dir=' + str(temp_output_dir))
+            objects = download_objects_inner(client, bucketname, parentdir, prefix_trunc,
+                    array_of_files, False, None, temp_input_dir, **kwargs)
+    if total_dirs_to_process == 0:
+        info('No directories to process')
+        return
+    else:
+        info("Total " + str(total_dirs_to_process) + " to process")
+        for debugdir in temp_input_roots:
+            print("DEBUG dir name " + debugdir + ",  exists="+str(os.path.exists(debugdir)))
+
+    if (enable_infin_ast == True):
+        transformAst = infin_ast.extract_transform(TRANSFORM_MULTI_INPUTS,\
+            src_str=transform_string)
+        transformSrc = infin_ast.get_source(transformAst)
+    else:
+        transformSrc = transform_string
+    xn = TRANSFORM_MULTI_INPUTS
+    tree = ast.parse(transformSrc)
+    compiledcode3 = compile(tree, "<string>", "exec")
+
+    # Add all functions in xformcode to the globals dictionary
+    glb = {}
+    fl = FuncLister(glb)
+    fl.visit(tree)
+    try:
+        exec(compiledcode3, glb)
+    except Exception as e:
+        status_str = str(e)
+        info("Execution of global statics failed. status_str=" + status_str)
+        delete_dirs(temp_input_roots)
+        raise
+    else:
+        status_str = 'Success'
+
+    # print('Globals=')
+    # pprint.pprint(glb)
+    info("Execution of global statics complete. status_str=" + status_str)
+
+    namespaced_infin_transform_fnx = namespaced_function(glb[xn], glb, None, True)
+    try:
+        namespaced_infin_transform_fnx(temp_input_roots, temp_output_dir, **kwargs)
+    except Exception as e:
+        status_str = str(e)
+        info("Error executing " + xn + ", status=" + status_str)
+    else:
+        info("Successfully executed transform " + xn)
+    finally:
+        delete_dirs(temp_input_roots)
+
+    try:
+        dest_path = 'infinstor/' + prefix_in[len(prefix_trunc):]
+        for one_output_file in os.listdir(temp_output_dir):
+            fq_local = os.path.join(temp_output_dir, one_output_file)
+            info("Logging artifact " + str(fq_local) + " to " + dest_path)
+            log_artifact(fq_local, dest_path)
+    except Exception as e:
+        status_str = str(e)
+        info("Error logging artifacts: " + status_str)
+    else:
+        info("Successfully logged artifacts")
+    finally:
+        shutil.rmtree(temp_output_dir)
+
+
 def run_transform_string_inline(client, bucketname, prefix_in, prefix_trunc,
         transform_string, **kwargs):
     if (look_for_transform(transform_string, TRANSFORM_RAW_PD)):
@@ -968,22 +1117,15 @@ def run_transform_string_inline(client, bucketname, prefix_in, prefix_trunc,
     else:
         run_script(transform_string, **kwargs)
 
-def run_transform_singlevm(service_name, parent_run_id, last_in_chain_of_xforms,
+def run_transform_singlevm(service_name, experiment_id, parent_run_id, last_in_chain_of_xforms,
         input_data_spec, xformname, instance_type, **kwargs):
     # save the conda environment
     projdir = tempfile.mkdtemp()
     if (verbose):
         print('Project dir: ' + projdir)
-    condayaml = os.open(projdir + sep + 'conda.yaml', os.O_CREAT|os.O_WRONLY, mode=0o644)
-    cmd = ['conda', 'env', 'export', '--no-builds']
-    if (verbose):
-        print('Running cmd:')
-        print(*cmd)
-    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,\
-                stdin=subprocess.DEVNULL, close_fds=True)
-    for line in process.stdout:
-        os.write(condayaml, line)
-    process.wait()
+    # create an empty conda.yaml file. mlflow requires it, but our backend does not use it
+    # The actual xform environment (docker or conda) is stored along with the xform
+    os.close(os.open(projdir + sep + 'conda.yaml', os.O_CREAT|os.O_WRONLY, mode=0o644))
 
     with open(projdir + sep + 'MLproject', "w") as projfile:
         kwp = ''
@@ -1019,8 +1161,9 @@ def run_transform_singlevm(service_name, parent_run_id, last_in_chain_of_xforms,
             + '", "last_in_chain_of_xforms": "' + last_in_chain_of_xforms + '"}'
 
     cmd = ['mlflow', 'run',
-            '-b', 'infinstor-backend', '--backend-config',
-            backend_config,
+            '-b', 'infinstor-backend',
+            '--backend-config', backend_config,
+            '--experiment-id', str(experiment_id),
             projdir,
             '-P', 'service=' + service_name,
             '-P', 'input_data_spec=' + json.dumps(input_data_spec),
@@ -1068,22 +1211,18 @@ def run_transform(run_options, **kwargs):
 
     service_name = run_options.pop('service_name')
     run_location = run_options.pop('run_location')
-        if (run_location == "inline"):
-        experiment_id = None
+    if (run_location == "inline"):
         if (parent_run_id != None):
             nes = True
-            if ('experiment_id' in run_options):
-                experiment_id = str(run_options['experiment_id'])
         else:
             nes = False
-        print("run_transform: experiment_id is " + experiment_id)
-        with start_run(nested=nes, experiment_id=experiment_id) as run:
+        with start_run(nested=nes, experiment_id=run_options['experiment_id']) as run:
             xobjects = run_transform_inline(service_name, run.info.run_id,
                     input_data_spec, xformname, **kwargs)
             return run.info.run_id
     elif (run_location == "singlevm"):
         instance_type = run_options.pop('instance_type')
-        return run_transform_singlevm(service_name, parent_run_id,
+        return run_transform_singlevm(service_name, run_options['experiment_id'], parent_run_id,
                 last_in_chain_of_xforms, input_data_spec, xformname, instance_type, **kwargs)
     elif (run_location == "emr"):
         time_spec = run_options.pop('time_spec')
@@ -1095,20 +1234,19 @@ def run_transform(run_options, **kwargs):
                 **kwargs)
 
 def run_transforms(transforms):
+    if ('MLFLOW_EXPERIMENT_ID' in os.environ):
+        experiment_id = os.environ['MLFLOW_EXPERIMENT_ID']
+    else:
+        experiment_id = '0'
     if (len(transforms) == 1):
         transform = transforms[0]
+        transform['run_options']['experiment_id'] = experiment_id
         if ('kwargs' in transform and transform['kwargs'] != None):
             return run_transform(transform['run_options'], **transform['kwargs'])
         else:
             return run_transform(transform['run_options'])
 
-    if ('experiment_id' in transforms[0]['run_options']):
-        experiment_id = str(transforms[0]['run_options']['experiment_id'])
-        print("run_transforms: experiment_id is specified as " + experiment_id)
-        run = start_run(experiment_id = experiment_id)
-    else:
-        print("run_transforms: experiment_id is not specified")
-        run = start_run()
+    run = start_run(experiment_id = experiment_id)
     parent_run_id = run.info.run_id
     print('parent_run_id = ' + str(parent_run_id))
 
@@ -1116,6 +1254,7 @@ def run_transforms(transforms):
     for ind in range(len(transforms)):
         transform = transforms[ind]
         run_options = transform['run_options']
+        run_options['experiment_id'] = experiment_id
 
         do_reset_mlflow_active_run_stack = False
         do_end_parent_run = False
@@ -1156,6 +1295,11 @@ def download_objects(client, bucketname, array_of_files, is_csv):
         return download_objects_inner(client, bucketname, '', '', array_of_files, is_csv,\
                 None, None)
 
+def delete_dirs(dirs):
+    for dir in dirs:
+        if os.path.exists(dir):
+            shutil.rmtree(dir)
+
 def test_infin_transform(service_name, input_data_spec, **kwargs):
     caller_globals = dict(inspect.getmembers(inspect.stack()[1][0]))["f_globals"]
     code_str = ''
